@@ -119,6 +119,13 @@ namespace Beagle.Util {
 				if (ea_support_tested)
 					return ea_supported;
 
+				if (Environment.GetEnvironmentVariable ("BEAGLE_DISABLE_XATTR") != null) {
+					Logger.Log.Debug ("BEAGLE_DISABLE_XATTR is set");
+					ea_supported = false;
+					ea_support_tested = true;
+					return false;
+				}
+
 				ea_supported = Test (PathFinder.HomeDir);
 				ea_support_tested = true;
 
